import serial, sys, time
from Tkinter import *
import tkFileDialog
import tkMessageBox
import tkFont
import threading

VERSION = "DL4195A v1.0"

ABOUT_TEXT = """
DL4195A: A simple GPIB data downloader for the HP4195A over
a Prologix GPIB/USB interface

(C) 2008 Tennessee Carmel-Veilleux

Permission is granted to use this program for any purpose, including
commercial as long as no fee is charged for use or distribution. No 
warranty, either express or implied is granted for the use of this
program. The author is not liable for any loss of data, damage or
injury arising from the use of this program.
"""

def float_conv(s):
    mult = 1.0
    
    if s[-1] == 'm':
        mult = 1e-3
        s = s[:-1]
    elif s[-1] == 'M':
        mult = 1e6
        s = s[:-1]
    elif s[-1] == 'u':
        mult = 1e-6
        s = s[:-1]
    elif s[-1] == 'p':
        mult = 1e-12
        s = s[:-1]
    elif s[-1] == 'n':
        mult = 1e-9
        s = s[:-1]
    elif s[-1] == 'K':
        mult = 1e3
        s = s[:-1]
    elif s[-1] == 'k':
        mult = 1e3
        s = s[:-1]
    
    return float(s) * mult

def scan():
    """scan for available ports. return a list of tuples (num, name)"""
    available = []
    for i in range(128):
        try:
            s = serial.Serial("COM%d" % ( i+1))
            available.append( (i, s.portstr))
            s.close()   #explicit close 'cause of delayed GC in java
        except serial.SerialException:
            pass
    return available
    
class BusyManager:

    def __init__(self, widget):
        self.toplevel = widget.winfo_toplevel()
        self.widgets = {}

    def busy(self, widget=None):

        # attach busy cursor to toplevel, plus all windows
        # that define their own cursor.

        if widget is None:
            w = self.toplevel # myself
        else:
            w = widget

        if not self.widgets.has_key(str(w)):
            try:
                # attach cursor to this widget
                cursor = w.cget("cursor")
                if cursor != "watch":
                    self.widgets[str(w)] = (w, cursor)
                    w.config(cursor="watch")
            except TclError:
                pass

        for w in w.children.values():
            self.busy(w)

    def notbusy(self):
        # restore cursors
        for w, cursor in self.widgets.values():
            try:
                w.config(cursor=cursor)
            except TclError:
                pass
        self.widgets = {}
        
class Dialog(Toplevel):
    def __init__(self, parent, title = None):
        Toplevel.__init__(self, parent)
        self.transient(parent)
        if title:
            self.title(title)
        
        self.parent = parent
        self.result = None

        body = Frame(self)
        self.initial_focus = self.body(body)
        body.pack(padx=5, pady=5)
        self.buttonbox()
        self.grab_set()
    
        if not self.initial_focus:
            self.initial_focus = self

        self.protocol("WM_DELETE_WINDOW", self.cancel)
        self.geometry("+%d+%d" % (parent.winfo_rootx()+50,parent.winfo_rooty()+50))
        self.initial_focus.focus_set()
        self.wait_window(self)
    #
    # construction hooks
    def body(self, master):
    # create dialog body. return widget that should have
    # initial focus. this method should be overridden
        pass
    
    def buttonbox(self):
    # add standard button box. override if you don't want the
    # standard buttons
        box = Frame(self)
        w = Button(box, text="OK", width=10, command=self.ok, default=ACTIVE)
        w.pack(side=LEFT, padx=5, pady=5)
        w = Button(box, text="Cancel", width=10, command=self.cancel)
        w.pack(side=LEFT, padx=5, pady=5)
        self.bind("<Return>", self.ok)
        self.bind("<Escape>", self.cancel)
        box.pack()
    #
    # standard button semantics
    def ok(self, event=None):
        if not self.validate():
            self.initial_focus.focus_set() # put focus back
            return
        self.withdraw()
        self.update_idletasks()
        self.apply()
        self.cancel()
        
    def cancel(self, event=None):
    # put focus back to the parent window
        self.parent.focus_set()
        self.destroy()
    
    #
    # command hooks
    def validate(self):
        return 1 # override
    
    def apply(self):
        pass # override

class TextViewer(Toplevel):
    """
    simple text viewer dialog for idle
    """
    def __init__(self, parent, title, fileName, data=None, xsize = 625, ysize = 500, font = None):
        """If data exists, load it into viewer, otherwise try to load file.

        fileName - string, should be an absolute filename
        """
        Toplevel.__init__(self, parent)
        self.configure(borderwidth=5)
        self.geometry("=%dx%d+%d+%d" % (xsize, ysize,
                                        parent.winfo_rootx() + 10,
                                        parent.winfo_rooty() + 10))
        #elguavas - config placeholders til config stuff completed
        self.bg = '#ffffff'
        self.fg = '#000000'

        self.font = font
        self.CreateWidgets()
        self.title(title)
        self.transient(parent)
        self.grab_set()
        self.protocol("WM_DELETE_WINDOW", self.Ok)
        self.parent = parent
        self.textView.focus_set()
        #key bindings for this dialog
        self.bind('<Return>',self.Ok) #dismiss dialog
        self.bind('<Escape>',self.Ok) #dismiss dialog
        if data:
            self.textView.insert(0.0, data)
        else:
            self.LoadTextFile(fileName)
        self.textView.config(state=DISABLED)
        self.wait_window()

    def LoadTextFile(self, fileName):
        textFile = None
        try:
            textFile = open(fileName, 'r')
        except IOError:
            tkMessageBox.showerror(title='File Load Error',
                    message='Unable to load file %r .' % (fileName,))
        else:
            self.textView.insert(0.0,textFile.read())

    def CreateWidgets(self):
        frameText = Frame(self, relief=SUNKEN, height=700)
        frameButtons = Frame(self)
        self.buttonOk = Button(frameButtons, text='Close',
                               command=self.Ok, takefocus=FALSE)
        self.scrollbarView = Scrollbar(frameText, orient=VERTICAL,
                                       takefocus=FALSE, highlightthickness=0)
                                       
        self.textView = Text(frameText, wrap=WORD, highlightthickness=0,
                             fg=self.fg, bg=self.bg)
        if self.font:
            self.textView.configure(font = self.font)
            
        self.scrollbarView.config(command=self.textView.yview)
        self.textView.config(yscrollcommand=self.scrollbarView.set)
        self.buttonOk.pack()
        self.scrollbarView.pack(side=RIGHT,fill=Y)
        self.textView.pack(side=LEFT,expand=TRUE,fill=BOTH)
        frameButtons.pack(side=BOTTOM,fill=X)
        frameText.pack(side=TOP,expand=TRUE,fill=BOTH)

    def Ok(self, event=None):
        self.destroy()

class Application(Frame):
    def __init__(self, comports, master=None):
        Frame.__init__(self, master)
        self.items = []

        self.comports = comports
        self.grid(padx = 10, pady = 10)
        self.createWidgets()
        #self.ed = EditorDialog(("Roboto","Patate","Chip   Chip "), self, title= "Edit Roboto", defaults = ("A","B",""))
        #print self.ed.results
        
    def getLog(self):
        return self.outputText
        
    def createWidgets(self):
        # Version title
        self.f1 = tkFont.Font(size = "10", weight = "bold")
        self.titleLabel = Label(self, font = self.f1, text = (VERSION + "\nBy Tennessee Carmel-Veilleux"))
        self.titleLabel.grid(columnspan = 3, row = 0, column = 0, pady = 5)
        
        # OptionMenu for COM PORT
        self.pFrame = Frame(self)
        self.l1 = Label(self.pFrame, text = "COM PORT:")
        self.l1.grid(row = 0, column = 0, padx = 5, pady = 5, sticky = W)
        
        if len(self.comports) > 0:
            default = self.comports[0]
        else:
            default = ""
            
        self.comportVar = StringVar()
        self.comportVar.set(default)
        #self.comportEntry = Entry(self.pFrame, width = 30, textvariable = self.comportVar)
        self.comportEntry = OptionMenu(self.pFrame, self.comportVar, *self.comports)
        self.comportEntry.grid(row = 0, column = 1, sticky = E+W)
        
        self.pFrame.grid(row = 1, column = 0, columnspan = 3, sticky = W)
        
        # Entry for GPIB address
        self.aFrame = Frame(self)
        self.l2 = Label(self.aFrame, text = "GPIB Address:")
        self.l2.grid(row = 0, column = 0, padx = 5, pady = 5, sticky = W)
        
        self.addrVar = StringVar()
        self.addrVar.set("17")
        
        self.addrEntry = Entry(self.aFrame, width = 10, textvariable = self.addrVar)
        self.addrEntry.grid(row = 0, column = 1, sticky = E+W)
        
        self.aFrame.grid(row = 2, column = 0, columnspan = 3, sticky = W)
        
        # Entry for filename
        self.fFrame = Frame(self)
        self.l3 = Label(self.fFrame, text = "Output file:")
        self.l3.grid(row = 0, column = 0, padx = 5, pady = 5, sticky = W)
        
        self.outputFilenameVar = StringVar()
        self.filenameEntry = Entry(self.fFrame, width = 50, textvariable = self.outputFilenameVar)
        self.filenameEntry.grid(row = 0, column = 1, sticky = E+W)
        self.outputFilenameVar.set("outfile.tsv")
        
        self.browseButton = Button(self.fFrame, text="...",command=self.onSelectFilename )
        self.browseButton.grid(column = 2, row = 0,  sticky = E+W)
        
        self.fFrame.grid(row = 3, column = 0, columnspan = 3, sticky = W)
        
        # Add execution buttons frame
        self.b2Frame = Frame(self)
        
        self.execButton = Button(self.b2Frame, text="Execute", command=self.onExecute)
        self.execButton.grid(column = 0, row = 0, padx = 10, sticky = E+W)
        
        self.aboutButton = Button(self.b2Frame, text="About...",command=self.onAbout )
        self.aboutButton.grid(column = 1, row = 0, padx = 10, sticky = E+W)
        
        
        self.quitButton = Button(self.b2Frame, text="Quit",command=self.onQuit )
        self.quitButton.grid(column = 2, row = 0, padx = 10, sticky = E+W)
        
        self.b2Frame.grid(columnspan = 3, row = 4, column = 0, sticky = E+W, pady = 10)
        self.bind("<Return>", self.onExecute)
        self.bind("<Escape>", self.onQuit)
        self.busyManager = BusyManager(self)

    def onSelectFilename(self, *args):
        filename = tkFileDialog.asksaveasfilename(filetypes=[("TSV files", "*.tsv")])
        if filename:
            self.outputFilenameVar.set(filename)
            
    def onAbout(self, *args):
        v = TextViewer(self, "About %s" % VERSION, None, ABOUT_TEXT, xsize = 400, ysize = 240)
    
    def onHelp(self, *args):
        #v = TextViewer(self, "Instructions for MergeBOM", None, HELP_TEXT, font = ("Courier",10))
        pass
        
    def onViewFile(self, *args):
        if len(self.lFrame.listBox.curselection()) > 0:
            selection = self.lFrame.listBox.curselection()[0]
            #print self.items[int(selection)]
            filename = self.items[int(selection)][1]
            v = TextViewer(self, "Showing " +  shortenFilename(filename, 50), filename, None)
    
    def validate(self):
        try:
            self.gpibAddr = int(self.addrVar.get())
            if not 0 <= self.gpibAddr <= 30:
                raise ValueError
        except ValueError:
            tkMessageBox.showerror(title='Error !', message="Address must be (0-30) !")
            self.addrVar.set("0")
            return False
        
        return True
            
    def onExecute(self, *args):
        if not self.validate():
            return
            
        self.busyManager.busy()
        self.execThread = threading.Thread(target = self.doExecute)
        self.killThread = threading.Timer(10.0, self.doTimeout)
        self.killThread.start()
        #self.execThread.start()
        self.doExecute()
        #self.execThread.join()
        self.busyManager.notbusy()
    
    def doExecute(self):
        try:
            # Get port number from first part of selection item
            comport = self.comportVar.get().split()[0]
            
            self.canceled = False
            
            # Get waveform
            output = self.gather_wave(comport, self.gpibAddr)
            
            # Cancel timeout if not triggered
            self.killThread.cancel()
            
            if self.canceled:
                tkMessageBox.showerror(title='Error !', message="Time-out while downloading data !")
                return
                
            if len(output) == 0:
                tkMessageBox.showerror(title='Error !', message="Problem while downloading data !")
                return
            
            # Write waveform to file
            filename = self.outputFilenameVar.get()
            f = open(filename, "w")
            for line in output:
                f.write(line)
            f.close()
            
            # Show a "done" message and the results
            tkMessageBox.showinfo(title='Information', message="Done saving data to\n'%s'" % filename)
            v = TextViewer(self, "Output file %s" % filename, filename, None)
        except Exception, m:
            tkMessageBox.showerror(title='Error !', message=str(m))
        
    def doTimeout(self):
        self.canceled = True
        
    def onQuit(self, *args):
        self.quit()
        
    def gather_wave(self, comport, gpibAddr):
        ser = serial.Serial()
        # Convert COMxxx to number
        ser.port  = serial.device(int(comport[3:]) - 1)
        ser.baudrate = 57600
        ser.rtscts  = False
        ser.xonxoff = False
        ser.timeout = 0.001
        try:
            ser.open()
        except serial.SerialException, e:
            raise Exception("Could not open serial port %s: %s" % (ser.portstr, e))

        ser.setTimeout(0.002)
        try:
            # Init instrument
            ser.write("++addr %d\r\n" % gpibAddr)
            ser.write("++eoi 0\r\n")
            ser.write("++eot_enable 1\r\n")
            ser.write("++eot_char 42\r\n")
            ser.write("++auto 1\r\n")
            ser.flushInput()

            ser.write("CPYM2\r\n")
            ser.write("COPY\r\n")
        except serial.SerialException:
            return []

        done = False
        data = ""

        # Read data
        while not done and not self.canceled:
            try:
                if ser.inWaiting() > 0:
                    buf = ser.read(256)
                    data = data + buf
                    
                    # End of opearation is marked by star
                    if buf.find("*") <> -1:
                        done = True
            except serial.SerialException:
                done = True

        # Reset copy mode to PRINT
        ser.write("CPYM3\r\n")

        # Return local control
        ser.write("++loc\r\n")

        # Flush buffer
        time.sleep(1.0)
        ser.flushInput()

        ser.close()

        # Extract lines
        lines = [s.strip() for s in data.strip('\0').strip().replace("*","").replace(chr(27),"").split('\n') if s.strip() <> "" and len(s.strip()) > 2]
        lines = [s for s in lines if len(s.strip()) > 2]

        output = []
        if len(lines) > 2:
            # Extract heading
            heading = lines[1]
            if heading[0] <> 'N':
                raise ValueError("First column of heading is not 'N': wrong data format")
            
            heading = [s for s in heading.split() if s <> ""]
            
            if len(heading) < 11:
                raise ValueError("Heading does not have proper format")
            
            if len(heading) == 13:
                heading =  "%s\t%s\t%s\t%s\n" % (heading[0],' '.join(heading[1:5]), ' '.join(heading[5:9]), ' '.join(heading[9:]))
            else:
                heading =  "%s\t%s\t%s\t%s\n" % (heading[0],' '.join(heading[1:5]), "A", "B")
            output.append(heading)
            
            # Extract line data
            for line in lines[2:]:
                n = int(line[0:5].strip())
                freq = float(line[5:22].strip().replace(" ",""))
                    
                fields = ["%e" % float_conv(s) for s in line[22:].split() if s <> ""]
                
                output.append("%d\t%f\t%s\n" % (n, freq, "\t".join(fields)))

        return output
        
if __name__ == "__main__":
    comports = ["COM%d" % (id+1) for id, name in scan()]
    
    app = Application(comports)
    app.master.title(VERSION)
    app.mainloop() 
    